val status = ref OS.Process.failure
fun quit() = OS.Process.exit (!status)


(* ---------- Error reporting functions ---------- *)

val errtup = ref NONE : (string * BasicIO.instream * Lexing.lexbuf) option ref
fun errPrint s = TextIO.output(TextIO.stdErr, s)

fun setErr (str, stream, buf) = errtup := SOME (str, stream, buf)

fun errPrint s = TextIO.output(TextIO.stdErr, s)

val verbose = ref false

fun vprint s =
    if !verbose then print s else ()

fun error l s =
    let val () = case (!errtup) of
		     NONE => errPrint ("Error: " ^ s ^ "\n")
		   | SOME x => (Location.errLocation x l;  Location.errPrompt (s^"\n"));
    in
	quit()
    end

fun exit s =
    let val () = errPrint ("Error: " ^ s ^ "\n")
    in quit ()
    end

fun warn l s =
    let
	val s' = "Warning: " ^ s ^ "\n"
    in
	case (!errtup) of
	    NONE => errPrint s'
	  | SOME x => (Location.errLocation x l;  Location.errPrompt s')
    end

fun ierror s =
    let
	val () = errPrint ("Internal error " ^ s ^ "\n")
    in
	quit()
    end

(* ---------------------------------------------------------------- *)

fun now() = Date.fmt "%H:%M:%S %Z, %A %d %B %Y" (Date.fromTimeLocal (Time.now()))

fun fst (x,y) = x
fun snd (x,y) = y

fun member _ [] = false
  | member x (h::t) = x=h orelse member x t;

fun truncate s n = String.substring (s , 0, size s - n)


(* ------------------------------------------------------------------------ *)

exception utilsError of string;

fun is_elem x xs = List.exists (fn y => x=y) xs

fun is_elem_with f x xs = List.exists (fn y => f(x,y)) xs

fun zipWith f [] [] = []
  | zipWith f (x::xs) (y::ys) = (f (x,y))::(zipWith f xs ys)
  | zipWith f _ _ = raise utilsError "zipWith: lists not of same length"

fun lookup x [] = NONE
  | lookup x ((y,z)::ys) = if (x=y) then SOME z else lookup x ys

fun splitBy x xs = String.tokens (fn c => c=x) xs;

fun substrAfterToken x xs =
    List.last (splitBy x xs);

fun substrBeforeToken x xs =
    List.hd (splitBy x xs);

(* perl-style join of a list *)
fun join c xs = foldr (fn (x,s) => x^(str c)^s) "" xs

fun list_diff xs ys = List.filter (fn x => not (is_elem x ys)) xs

fun list_diff_with f xs ys = List.filter (fn x => not (is_elem_with f x ys)) xs

fun zip ([],[]) = []
  | zip (h1::t1,h2::t2) = (h1,h2) :: (zip (t1,t2))
  | zip _  = raise utilsError "zip: lists not of same length"

fun fst (x,y) = x
fun snd (x,y) = y
fun id x = x

(*stolen from Camelot/src/Util.sml*)
fun listToString toString separator l =
case l of [] => ""
        | [h] => toString h
        | h::t => (toString h) ^ separator ^ (listToString toString separator t)

(* same as Util.makeFullFilename; need here to be standalone *)
fun makeFullFilename dir base ext =
    Path.joinDirFile
       {dir=dir,
	file=Path.joinBaseExt {base=base, ext=SOME ext}}

fun basename f = List.last (String.tokens (fn c => ( c = #"/")) f)

