(* Real32.sml
 *
 * Peter Bertelsen
 * August 1997
 *)

open Word8Vector

datatype real = REAL of vector

fun fromReal r = REAL (Mosml.floatVec r)

fun toReal (REAL bs) = Mosml.vecFloat bs
(* Previously unimplemented.  This gives some unpleasant rounding
   errors.  For instance,  0.9 comes out as 0.899999976158.
   The 64-bit version is much better.*)

fun toString x = Real.toString (toReal x)

fun fromBytes bs = if length bs = 4 then SOME (REAL bs)
		   else NONE

fun toBytes (REAL bs) = bs

fun emit out (REAL bs) = app out bs

fun scan src =
    (let val b3 = valOf(src())
	 val b2 = valOf(src())
	 val b1 = valOf(src())
	 val b0 = valOf(src())
     in
	 SOME (REAL (fromList[b3, b2, b1, b0]))
     end)
	 handle Option => NONE


(* kwxm: convert a hex representation like "0fx874fe155" to Real32 and back *)

exception Real32HexError of string

fun isHexString s = List.all Char.isHexDigit (explode s)

fun get2 s n = valOf (Word8.fromString (String.extract (s,2*n,SOME 2)))
    handle Option => raise Real32HexError "This can't happen"

fun fromHexString t =
    if not (size t = 8) then raise Real32HexError ("Bad length: 8 digits expected")
    else if not (isHexString t) then raise Real32HexError "Bad hex digit"
    else
	let
	    val l = List.tabulate (4, get2 t)
	in
	    valOf (fromBytes (Word8Vector.fromList l))
	end

fun tohex2 n = let
    val s = Word8.toString n
in if n < 0w16 then "0" ^ s else s end

fun toHexString (REAL l) = String.map Char.toLower
				      (foldl (fn (x,s) => s ^ (tohex2 x)) "" l)

